defmodule ModuleGeneratorMacros do

  @moduledoc """
  """

# An alternative definition of gen_dispatch_table that does not use
# the bind_quoted option.
#
#   defmacro gen_dispatch_table() do
#     methods = ModuleGeneratorHelpers.get_methods()
#     ast = for item <- methods do
#       quote do
#         {unquote(String.to_atom(:proplists.get_value("method_name", item))),
#           unquote(String.to_atom(:proplists.get_value("method_name", item)))}
#       end
#     end
#     #IO.inspect(ast, label: "ast")
#     ast
#   end

  defmacro gen_dispatch_table() do
    methods = ModuleGeneratorHelpers.get_methods()
    ast = for item <- methods do
      quote bind_quoted: [method_name: String.to_atom(:proplists.get_value("method_name", item))] do
        {method_name, method_name}
      end
    end
    #IO.inspect(ast, label: "ast")
    ast
  end

end

defmodule ModuleGeneratorHelpers do
  @moduledoc """
  """

  {:ok, spec_txt} = File.read("Data/module03.exs")
  {spec, _} = Code.eval_string(spec_txt)
  @methods :proplists.get_value("instance_methods", spec)

  def get_methods() do
    @methods
  end

  def test1(pid_in, opts \\ []) do
    pid = if is_nil(pid_in) do
      {:ok, pid} = ModuleGenerator.start([123, 456])
      pid
    else
      pid_in
    end
    result = ModuleGenerator.rpc(pid, :methodName2, [111, 222, 333],
      opt1: :aa, opt2: :bb)
    IO.inspect(result, label: "result")
    result = ModuleGenerator.rpc(pid, :methodName4, [111, 222, 333],
      opt1: :aa, opt2: :bb)
    IO.inspect(result, label: "result")
    result = ModuleGenerator.rpc(pid, :methodName8, [111, 222, 333],
      opt1: :aa, opt2: :bb)
    IO.inspect(result, label: "result")
    result = ModuleGenerator.rpc(pid, :methodName2, [111, 222, 333],
      opt1: :aa, opt2: :bb)
    IO.inspect(result, label: "result")
    result = ModuleGenerator.rpc(pid, :methodName4, [111, 222, 333],
      opt1: :aa, opt2: :bb)
    IO.inspect(result, label: "result")
    if Keyword.has_key?(opts, :kill) and Keyword.get(opts, :kill) do
      ModuleGenerator.rpc(pid, :exit, nil, nil)
    end
    # maybe this process is no longer alive.
    pid
  end

  def test2(pid_in, opts \\ []) do
    pid = if is_nil(pid_in) do
      {:ok, pid} = ModuleGenerator.start([123, 456])
      pid
    else
      pid_in
    end
    ModuleGeneratorHelpers.get_methods()
    |> Enum.each(fn (item) ->
      method_name = :proplists.get_value("method_name", item)
      method_name_atom = String.to_atom(method_name)
      result = ModuleGenerator.rpc(pid, method_name_atom, [111, 222, 333],
        opt3: :aa, opt3: :bb)
      IO.inspect(result, label: "result")
    end)
    if Keyword.has_key?(opts, :kill) and Keyword.get(opts, :kill) do
      ModuleGenerator.rpc(pid, :exit, nil, nil)
    end
    # maybe this process is no longer alive.
    pid
  end

  @doc """
  Repeat test2 count number of times.
  """
  @spec repeat_test2(pid, Integer.t()) :: any
  def repeat_test2(pid, count) do
    1 .. count
    |> Enum.each(fn (n) ->
      IO.puts("\ncount: #{n}")
      ModuleGeneratorHelpers.test2 pid
    end)
  end

  @doc """
  Read a yaml spec file.  Write out the equivalent string
  representation of the data structure returned by `fast_yaml`.  By
  doing this, we avoid the need for `fast_yaml` at compile-time, which
  causes a `nif` not loaded error.

  ## Options

    * `:force` - if true, force over-write of target file.

  ## Example

      iex> ModuleGeneratorHelpers.save_yaml_spec_to_text_file("spec.yaml", "spec.txt")

  """
  @spec save_yaml_spec_to_text_file(String.t(), String.t(), Keyword.t()) ::
      :ok | {:error, atom()}
  def save_yaml_spec_to_text_file(in_spec_file, out_spec_text_file, opts \\ []) do
    force =  Keyword.has_key?(opts, :force) and Keyword.get(opts, :force)
    if not force and File.exists?(out_spec_text_file) do
      msg = "error.  file \"#{out_spec_text_file}\" exists.  use force: true to overwrite"
      IO.puts(:stderr, msg)
      {:error, :file_exists, msg}
    else
      reply = :fast_yaml.decode_from_file(in_spec_file)
      case reply do
        {:ok, [spec | _]} ->
          spec_text = Kernel.inspect(spec)
          File.write(out_spec_text_file, spec_text)
          :ok
        {:error, reason} ->
          msg =  "can't read file \"#{in_spec_file}\".  reason: #{reason}"
          IO.puts(:stderr, msg)
          {:error, reason, msg}
      end
    end
  end

  def show_method_bodies() do
    methods = ModuleGeneratorHelpers.get_methods()
    methods |> Enum.each(fn (item) ->
      name = :proplists.get_value("method_name", item)
      body = :proplists.get_value("method_body", item)
      IO.puts("name: #{name}  body:\n#{body}")
    end)
  end

end

defmodule ModuleGenerator do
  @moduledoc """
  """

  require ModuleGeneratorMacros

  @doc """
  Create and return the process that implements the module and its functions.

  ## Example

      iex> {:ok, pid} = ModuleGenerator.start([123, 456])

  """
  @spec start(any) :: {:ok, pid}
  def start(data) do
    pid = spawn(__MODULE__, :module01_proc, [data])
    {:ok, pid}
  end

  @doc """
  Handle query requests.

  ## Example

      iex> ModuleGenerator.rpc(pid, :methodName2, [111, 222], [333, 444])
      
  """
  @spec rpc(pid, atom, [any], [any]) :: {:ok, any} | {:error, String.t()}
  def rpc(pid, method, args, opts) do
    if not Process.alive?(pid) do
      IO.puts("process is not alive")
      {:error, "process not alive"}
    else
      send(pid, {method, args, opts, self()})
      receive do
        result ->
          result
      after 2000 ->
          msg = "timed out"
          IO.puts(msg)
          {:error, msg}
      end
    end
  end

  # Generate the dispatch table
  # You can use:
  #     iex> ModuleGenerator.get_dispatch_table
  # to view this table.
  @dispatch_table ModuleGeneratorMacros.gen_dispatch_table()

  def get_dispatch_table() do
    @dispatch_table
  end

  @doc """
  A function that implements the process that dispatches to
  generated functions.  The argument can be any data structure.
  That value is passed to each dispatched function, saved for
  future dispatches.

  Compare this with `GenServer.handle_call/3` and `GenServer.handle_cast/2`.

  It is not intended that you call this directly.  Consider instead,
  calling it through `rpc/4`.
  """
  @spec module01_proc(any) :: any
  def module01_proc(data) do
    receive do
      {:exit, _, _, pid} ->
        send( pid, :ok)
      {method, args, opts, pid} ->
        IO.puts("(module01_proc) got message")
        dispatch_table = get_dispatch_table()
        # use the dispatch table to determine whether a function has
        # been generated for this atom/name
        if Keyword.has_key?(dispatch_table, method) do
          method_atom = Keyword.get(dispatch_table, method)
          response = apply(__MODULE__, method_atom, [args, opts]) 
          case response do
            {:ok, result} ->
              send(pid, {:ok, result})
              module01_proc(data)
            _ ->
              send(pid, {:error, "func not found"})
              module01_proc(data)
          end
        else
            IO.puts("function #{Atom.to_string(method)} does not exist")
            send(pid, {:error, "function does not exist"})
            module01_proc(data)
        end
    end
  end

  ModuleGeneratorHelpers.get_methods()
  |> Enum.each(fn (item) ->
    method_name = :proplists.get_value("method_name", item)
    method_name_atom = String.to_atom(method_name)
    method_body_str = :proplists.get_value("method_body", item)
    {:ok, method_body_ast} = Code.string_to_quoted(method_body_str)
    #IO.inspect(method_body_ast, label: "method_body_ast")
    def unquote(method_name_atom)(args, opts) do
      unquote(method_body_ast)
    end
  end)

end

